from abc import ABC
from typing import Any

import requests

from .config import Config


class Resource(ABC, dict):
    def __init__(self, config: Config, session: requests.Session, **kwargs):
        super().__init__(**kwargs)
        self.config: Config
        super().__setattr__("config", config)
        self.session: requests.Session
        super().__setattr__("session", session)

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError("cannot set attributes: use update() instead")


class Resources(ABC):
    def __init__(self, config: Config, session: requests.Session) -> None:
        self.config = config
        self.session = session



import requests
import responses

from responses import matchers

from posit.connect.client import Client
from posit.connect.config import Config
from posit.connect.content import ContentItem, ContentItemOwner
from posit.connect.permissions import Permissions

from .api import load_mock  # type: ignore


class TestContentOwnerAttributes:
    @classmethod
    def setup_class(cls):
        guid = "20a79ce3-6e87-4522-9faf-be24228800a4"
        config = Config(api_key="12345", url="https://connect.example/")
        session = requests.Session()
        fake_item = load_mock(f"v1/users/{guid}.json")
        cls.item = ContentItemOwner(config, session, **fake_item)

    def test_guid(self):
        assert self.item.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

    def test_username(self):
        assert self.item.username == "carlos12"

    def test_first_name(self):
        assert self.item.first_name == "Carlos"

    def test_last_name(self):
        assert self.item.last_name == "User"


class TestContentItemAttributes:
    @classmethod
    def setup_class(cls):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        config = Config(api_key="12345", url="https://connect.example/")
        session = requests.Session()
        fake_item = load_mock(f"v1/content/{guid}.json")
        cls.item = ContentItem(config, session, **fake_item)

    def test_id(self):
        assert self.item.id == "8274"

    def test_guid(self):
        assert self.item.guid == "f2f37341-e21d-3d80-c698-a935ad614066"

    def test_name(self):
        assert self.item.name == "Performance-Data-1671216053560"

    def test_title(self):
        assert self.item.title == "Performance Data"

    def test_description(self):
        assert self.item.description == ""

    def test_access_type(self):
        assert self.item.access_type == "logged_in"

    def test_connection_timeout(self):
        assert self.item.connection_timeout is None

    def test_read_timeout(self):
        assert self.item.read_timeout is None

    def test_init_timeout(self):
        assert self.item.init_timeout is None

    def test_idle_timeout(self):
        assert self.item.idle_timeout is None

    def test_max_processes(self):
        assert self.item.max_processes is None

    def test_min_processes(self):
        assert self.item.min_processes is None

    def test_max_conns_per_process(self):
        assert self.item.max_conns_per_process is None

    def test_load_factor(self):
        assert self.item.load_factor is None

    def test_cpu_request(self):
        assert self.item.cpu_request is None

    def test_cpu_limit(self):
        assert self.item.cpu_limit is None

    def test_memory_request(self):
        assert self.item.memory_request is None

    def test_memory_limit(self):
        assert self.item.memory_limit is None

    def test_amd_gpu_limit(self):
        assert self.item.amd_gpu_limit is None

    def test_nvidia_gpu_limit(self):
        assert self.item.nvidia_gpu_limit is None

    def test_created_time(self):
        assert self.item.created_time == "2022-12-16T18:40:53Z"

    def test_last_deployed_time(self):
        assert self.item.last_deployed_time == "2024-02-24T09:56:30Z"

    def test_bundle_id(self):
        assert self.item.bundle_id == "401171"

    def test_app_mode(self):
        assert self.item.app_mode == "quarto-static"

    def test_content_category(self):
        assert self.item.content_category == ""

    def test_parameterized(self):
        assert self.item.parameterized is False

    def test_cluster_name(self):
        assert self.item.cluster_name == "Local"

    def test_image_name(self):
        assert self.item.image_name is None

    def test_default_image_name(self):
        assert self.item.default_image_name is None

    def test_default_r_environment_management(self):
        assert self.item.default_r_environment_management is None

    def test_default_py_environment_management(self):
        assert self.item.default_py_environment_management is None

    def test_service_account_name(self):
        assert self.item.service_account_name is None

    def test_r_version(self):
        assert self.item.r_version is None

    def test_r_environment_management(self):
        assert self.item.r_environment_management is None

    def test_py_version(self):
        assert self.item.py_version == "3.9.17"

    def test_py_environment_management(self):
        assert self.item.py_environment_management is True

    def test_quarto_version(self):
        assert self.item.quarto_version == "1.3.340"

    def test_run_as(self):
        assert self.item.run_as is None

    def test_run_as_current_user(self):
        assert self.item.run_as_current_user is False

    def test_owner_guid(self):
        assert self.item.owner_guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

    def test_content_url(self):
        assert (
            self.item.content_url
            == "https://connect.example/content/f2f37341-e21d-3d80-c698-a935ad614066/"
        )

    def test_dashboard_url(self):
        assert (
            self.item.dashboard_url
            == "https://connect.example/connect/#/apps/f2f37341-e21d-3d80-c698-a935ad614066"
        )

    def test_app_role(self):
        assert self.item.app_role == "viewer"

    def test_owner(self):
        assert "owner" not in self.item

    def test_permissions(self):
        assert isinstance(self.item.permissions, Permissions)

    def test_tags(self):
        assert self.item.tags == []


class TestContentItemGetContentOwner:
    @responses.activate
    def test_owner(self):
        mock_content = load_mock(
            "v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"
        )
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=mock_content,
        )

        mock_user_get = responses.get(
            f"https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock(
                f"v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"
            ),
        )

        c = Client("12345", "https://connect.example")
        item = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        owner = item.owner
        assert owner.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

        # load a second time, assert tha owner is loaded from cached result
        owner = item.owner
        assert owner.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"
        assert mock_user_get.call_count == 1


class TestContentItemDelete:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{guid}"
        )

        # setup
        config = Config(api_key="12345", url="https://connect.example/")
        session = requests.Session()
        fake_item = load_mock(f"v1/content/{guid}.json")
        item = ContentItem(config, session, **fake_item)

        # invoke
        item.delete()

        # assert
        assert mock_delete.call_count == 1


class TestContentDeploy:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        task_id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_content_deploy = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/deploy",
            match=[matchers.json_params_matcher({"bundle_id": None})],
            json={"task_id": task_id},
        )

        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{task_id}",
            json=load_mock(f"v1/tasks/{task_id}.json"),
        )

        # setup
        c = Client("12345", "https://connect.example")
        content = c.content.get(content_guid)

        # invoke
        task = content.deploy()

        # assert
        task.id == task_id
        assert mock_content_get.call_count == 1
        assert mock_content_deploy.call_count == 1
        assert mock_tasks_get.call_count == 1


class TestContentUpdate:
    @responses.activate
    def test_update(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )
        con = Client("12345", "https://connect.example")
        content = con.content.get(guid)
        assert content.guid == guid

        new_name = "New Name"
        fake_content = load_mock(f"v1/content/{guid}.json")
        fake_content.update(name=new_name)
        responses.patch(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=fake_content,
        )

        content.update(name=new_name)
        assert content.name == new_name


class TestContentCreate:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_content_item = load_mock(f"v1/content/{guid}.json")

        # behavior
        responses.post(
            f"https://connect.example/__api__/v1/content",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    {"name": fake_content_item["name"]}
                )
            ],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example/")

        # invoke
        content_item = client.content.create(name=fake_content_item["name"])

        # assert
        assert content_item.name == fake_content_item["name"]


class TestContentsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "owner,tags"})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        content = client.content.find()

        #  assert
        assert mock_get.call_count == 1
        assert len(content) == 3
        assert content[0].name == "team-admin-dashboard"
        assert content[1].name == "Performance-Data-1671216053560"
        assert content[2].name == "My-Streamlit-app"

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        client.content.find(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        client.content.find(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "owner,tags"})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        content_item = client.content.find_one()

        #  assert
        assert mock_get.call_count == 1
        assert content_item

    @responses.activate
    def test_owner_guid(self):
        # data
        owner_guid = "a01792e3-2e67-402e-99af-be04a48da074"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[
                matchers.query_param_matcher(
                    {"owner_guid": owner_guid, "include": "owner,tags"}
                )
            ],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        content_item = client.content.find_one(owner_guid=owner_guid)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item.owner_guid == owner_guid

    @responses.activate
    def test_name(self):
        # data
        name = "team-admin-dashboard"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[
                matchers.query_param_matcher(
                    {"name": name, "include": "owner,tags"}
                )
            ],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        content_item = client.content.find_one(name=name)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item.name == name

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        client.content.find_one(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("12345", "https://connect.example")

        # invoke
        client.content.find_one(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsGet:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock(
                "v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"
            ),
        )
        con = Client("12345", "https://connect.example")
        get_one = con.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        assert get_one.name == "Performance-Data-1671216053560"


class TestContentsCount:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.content.count()
        assert count == 3



"""Metric resources."""

from .. import resources

from . import usage
from .usage import Usage


class Metrics(resources.Resources):
    """Metrics resource.

    Attributes
    ----------
    usage: Usage
        Usage resource.
    """

    @property
    def usage(self) -> Usage:
        return Usage(self.config, self.session)



import warnings

from http.client import responses
from requests import JSONDecodeError, Response

from .errors import ClientError


def handle_errors(response: Response, *args, **kwargs) -> Response:
    if response.status_code >= 400:
        try:
            data = response.json()
            error_code = data["code"]
            message = data["error"]
            http_status = response.status_code
            http_status_message = responses[http_status]
            raise ClientError(
                error_code, message, http_status, http_status_message
            )
        except JSONDecodeError:
            # No JSON error message from Connect, so just raise
            response.raise_for_status()
    return response


def check_for_deprecation_header(
    response: Response, *args, **kwargs
) -> Response:
    """
    Check for deprecation warnings from the server.

    You might get these if you've upgraded the Connect server but not posit-sdk.
    posit-sdk will make the right request based on the version of the server,
    but if you have an old version of the package, it won't know the new URL
    to request.
    """
    if "X-Deprecated-Endpoint" in response.headers:
        msg = (
            response.url
            + " is deprecated and will be removed in a future version of Connect."
            + " Please upgrade `posit-sdk` in order to use the new APIs."
        )
        warnings.warn(msg, DeprecationWarning)
    return response



import pytest

from posit.connect import urls


class TestCreate:
    def test(self):
        url = "http://example.com/__api__"
        assert urls.create(url) == url

    def test_append_path(self):
        assert (
            urls.create("http://example.com/") == "http://example.com/__api__"
        )

    def test_missing_scheme(self):
        with pytest.raises(ValueError):
            urls.create("example.com")

    def test_missing_netloc(self):
        with pytest.raises(ValueError):
            urls.create("http://")


class TestAppend:
    def test(self):
        url = "http://example.com/__api__"
        url = urls.create(url)
        assert urls.append(url, "path") == "http://example.com/__api__/path"

    def test_slash_prefix(self):
        url = "http://example.com/__api__"
        url = urls.create(url)
        assert urls.append(url, "/path") == "http://example.com/__api__/path"

    def test_slash_suffix(self):
        url = "http://example.com/__api__"
        url = urls.create(url)
        assert urls.append(url, "path/") == "http://example.com/__api__/path"



"""Client connection for Posit Connect."""

from __future__ import annotations

from requests import Response, Session
from typing import Optional

from . import hooks, me, urls

from .auth import Auth
from .config import Config
from .oauth import OAuthIntegration
from .content import Content
from .metrics import Metrics
from .tasks import Tasks
from .users import User, Users
from .groups import Groups


class Client:
    """
    Client connection for Posit Connect.

    This class provides an interface to interact with the Posit Connect API,
    allowing for authentication, resource management, and data retrieval.

    Parameters
    ----------
    api_key : str, optional
        API key for authentication
    url : str, optional
        Sever API URL

    Attributes
    ----------
    content: Content
        Content resource.
    me: User
        Connect user resource.
    metrics: Metrics
        Metrics resource.
    tasks: Tasks
        Tasks resource.
    users: Users
        Users resource.
    version: str
        Server version.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        self.config = Config(api_key=api_key, url=url)
        session = Session()
        session.auth = Auth(config=self.config)
        session.hooks["response"].append(hooks.check_for_deprecation_header)
        session.hooks["response"].append(hooks.handle_errors)
        self.session = session

    @property
    def version(self) -> str:
        """
        The server version.

        Returns
        -------
        str
            The version of the Posit Connect server.
        """
        return self.get("server_settings").json()["version"]

    @property
    def me(self) -> User:
        """
        The connected user.

        Returns
        -------
        User
            The currently authenticated user.
        """
        return me.get(self.config, self.session)

    @property
    def oauth(self) -> OAuthIntegration:
        """
        An OAuthIntegration.

        Returns
        -------
        OAuthIntegration
            The OAuth integration instance.
        """
        return OAuthIntegration(config=self.config, session=self.session)

    @property
    def groups(self) -> Groups:
        """The groups resource interface.

        Returns
        -------
        Groups
            The groups resource interface.
        """
        return Groups(self.config, self.session)

    @property
    def tasks(self) -> Tasks:
        """
        The tasks resource interface.

        Returns
        -------
        tasks.Tasks
            The tasks resource instance.
        """
        return Tasks(self.config, self.session)

    @property
    def users(self) -> Users:
        """
        The users resource interface.

        Returns
        -------
        Users
            The users resource instance.
        """
        return Users(config=self.config, session=self.session)

    @property
    def content(self) -> Content:
        """
        The content resource interface.

        Returns
        -------
        Content
            The content resource instance.
        """
        return Content(config=self.config, session=self.session)

    @property
    def metrics(self) -> Metrics:
        """
        The Metrics API interface.

        The Metrics API is designed for capturing, retrieving, and managing
        quantitative measurements of Connect interactions. It is commonly used
        for monitoring and analyzing system performance, user behavior, and
        business processes. This API facilitates real-time data collection and
        accessibility, enabling organizations to make informed decisions based
        on key performance indicators (KPIs).

        Returns
        -------
        Metrics
            The metrics API instance.

        Examples
        --------
        >>> from posit import connect
        >>> client = connect.Client()
        >>> content_guid = "2243770d-ace0-4782-87f9-fe2aeca14fc8"
        >>> events = client.metrics.usage.find(content_guid=content_guid)
        >>> len(events)
        24
        """
        return Metrics(self.config, self.session)

    def __del__(self):
        """Close the session when the Client instance is deleted."""
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def __enter__(self):
        """Enter method for using the client as a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Close the session if it exists.

        Parameters
        ----------
        exc_type : type
            The type of the exception raised (if any).
        exc_value : Exception
            The exception instance raised (if any).
        exc_tb : traceback
            The traceback for the exception raised (if any).
        """
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def request(self, method: str, path: str, **kwargs) -> Response:
        """
        Send an HTTP request.

        A facade for [](`requests.request`) configured for the target server.

        Parameters
        ----------
        method : str
            The HTTP method to use for the request.
        path : str
            Appended to the url object attribute.
        **kwargs
            Additional keyword arguments passed to [](`requests.request`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.request(method, url, **kwargs)

    def get(self, path: str, **kwargs) -> Response:
        """
        Send a GET request.

        A facade for [](`requests.get`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.get`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.get(url, **kwargs)

    def post(self, path: str, **kwargs) -> Response:
        """
        Send a POST request.

        A facade for [](`requests.post`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.post`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.post(url, **kwargs)

    def put(self, path: str, **kwargs) -> Response:
        """
        Send a PUT request.

        A facade for [](`requests.put`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.put`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.put(url, **kwargs)

    def patch(self, path: str, **kwargs) -> Response:
        """
        Send a PATCH request.

        A facade for [](`requests.patch`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.patch`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.patch(url, **kwargs)

    def delete(self, path: str, **kwargs) -> Response:
        """
        Send a DELETE request.

        A facade for [](`requests.delete`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.delete`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = urls.append(self.config.url, path)
        return self.session.delete(url, **kwargs)



"""Bundle resources."""

from __future__ import annotations

import io
import requests
from typing import List
from . import config, resources, tasks, urls


class BundleMetadata(resources.Resource):
    """Bundle metadata resource.

    Attributes
    ----------
    source : str | None
        Source of the bundle.
    source_repo : str | None
        Source repository of the bundle.
    source_branch : str | None
        Source branch of the bundle.
    source_commit : str | None
        Source commit of the bundle.
    archive_md5 : str | None
        MD5 checksum of the bundle archive.
    archive_sha1 : str | None
        SHA-1 checksum of the bundle archive.
    """

    @property
    def source(self) -> str | None:
        return self.get("source")

    @property
    def source_repo(self) -> str | None:
        return self.get("source_repo")

    @property
    def source_branch(self) -> str | None:
        return self.get("source_branch")

    @property
    def source_commit(self) -> str | None:
        return self.get("source_commit")

    @property
    def archive_md5(self) -> str | None:
        return self.get("archive_md5")

    @property
    def archive_sha1(self) -> str | None:
        return self.get("archive_sha1")


class Bundle(resources.Resource):
    """Bundle resource.

    Attributes
    ----------
    id : str
        Identifier of the bundle.
    content_guid : str
        Content GUID of the bundle.
    created_time : str
        Creation time of the bundle.
    cluster_name : str | None
        Cluster name associated with the bundle.
    image_name : str | None
        Image name of the bundle.
    r_version : str | None
        R version used in the bundle.
    r_environment_management : bool | None
        Indicates if R environment management is enabled.
    py_version : str | None
        Python version used in the bundle.
    py_environment_management : bool | None
        Indicates if Python environment management is enabled.
    quarto_version : str | None
        Quarto version used in the bundle.
    active : bool | None
        Indicates if the bundle is active.
    size : int | None
        Size of the bundle.
    metadata : BundleMetadata
        Metadata of the bundle.
    """

    @property
    def id(self) -> str:
        return self["id"]

    @property
    def content_guid(self) -> str:
        return self["content_guid"]

    @property
    def created_time(self) -> str:
        return self["created_time"]

    @property
    def cluster_name(self) -> str | None:
        return self.get("cluster_name")

    @property
    def image_name(self) -> str | None:
        return self.get("image_name")

    @property
    def r_version(self) -> str | None:
        return self.get("r_version")

    @property
    def r_environment_management(self) -> bool | None:
        return self.get("r_environment_management")

    @property
    def py_version(self) -> str | None:
        return self.get("py_version")

    @property
    def py_environment_management(self) -> bool | None:
        return self.get("py_environment_management")

    @property
    def quarto_version(self) -> str | None:
        return self.get("quarto_version")

    @property
    def active(self) -> bool | None:
        return self["active"]

    @property
    def size(self) -> int | None:
        return self["size"]

    @property
    def metadata(self) -> BundleMetadata:
        return BundleMetadata(
            self.config, self.session, **self.get("metadata", {})
        )

    def delete(self) -> None:
        """Delete the bundle."""
        path = f"v1/content/{self.content_guid}/bundles/{self.id}"
        url = urls.append(self.config.url, path)
        self.session.delete(url)

    def deploy(self) -> tasks.Task:
        """Deploy the bundle.

        Spawns an asynchronous task, which activates the bundle.

        Returns
        -------
        tasks.Task
            The task for the deployment.

        Examples
        --------
        >>> task = bundle.deploy()
        >>> task.wait_for()
        None
        """
        path = f"v1/content/{self.content_guid}/deploy"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, json={"bundle_id": self.id})
        result = response.json()
        ts = tasks.Tasks(self.config, self.session)
        return ts.get(result["task_id"])

    def download(self, output: io.BufferedWriter | str) -> None:
        """Download a bundle.

        Download a bundle to a file or memory.

        Parameters
        ----------
        output : io.BufferedWriter or str
            An io.BufferedWriter instance or a str representing a relative or absolute path.

        Raises
        ------
        TypeError
            If the output is not of type `io.BufferedWriter` or `str`.

        Examples
        --------
        Write to a file.
        >>> bundle.download("bundle.tar.gz")
        None

        Write to an io.BufferedWriter.
        >>> with open('bundle.tar.gz', 'wb') as file:
        >>>     bundle.download(file)
        None
        """
        if not isinstance(output, (io.BufferedWriter, str)):
            raise TypeError(
                f"download() expected argument type 'io.BufferedWriter` or 'str', but got '{type(output).__name__}'"
            )

        path = f"v1/content/{self.content_guid}/bundles/{self.id}/download"
        url = urls.append(self.config.url, path)
        response = self.session.get(url, stream=True)
        if isinstance(output, io.BufferedWriter):
            for chunk in response.iter_content():
                output.write(chunk)
        elif isinstance(output, str):
            with open(output, "wb") as file:
                for chunk in response.iter_content():
                    file.write(chunk)


class Bundles(resources.Resources):
    """Bundles resource.

    Parameters
    ----------
    config : config.Config
        Configuration object.
    session : requests.Session
        HTTP session object.
    content_guid : str
        Content GUID associated with the bundles.

    Attributes
    ----------
    content_guid: str
        Content GUID associated with the bundles.
    """

    def __init__(
        self,
        config: config.Config,
        session: requests.Session,
        content_guid: str,
    ) -> None:
        super().__init__(config, session)
        self.content_guid = content_guid

    def create(self, input: io.BufferedReader | bytes | str) -> Bundle:
        """
        Create a bundle.

        Create a bundle from a file or memory.

        Parameters
        ----------
        input : io.BufferedReader, bytes, or str
            Input archive for bundle creation. A 'str' type assumes a relative or absolute filepath.

        Returns
        -------
        Bundle
            The created bundle.

        Raises
        ------
        TypeError
            If the input is not of type `io.BufferedReader`, `bytes`, or `str`.

        Examples
        --------
        Create a bundle from io.BufferedReader
        >>> with open('bundle.tar.gz', 'rb') as file:
        >>>     bundle.create(file)
        None

        Create a bundle from bytes.
        >>> with open('bundle.tar.gz', 'rb') as file:
        >>>     data: bytes = file.read()
        >>>     bundle.create(data)
        None

        Create a bundle from pathname.
        >>> bundle.create("bundle.tar.gz")
        None
        """
        if isinstance(input, (io.BufferedReader, bytes)):
            data = input
        elif isinstance(input, str):
            with open(input, "rb") as file:
                data = file.read()
        else:
            raise TypeError(
                f"create() expected argument type 'io.BufferedReader', 'bytes', or 'str', but got '{type(input).__name__}'"
            )

        path = f"v1/content/{self.content_guid}/bundles"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, data=data)
        result = response.json()
        return Bundle(self.config, self.session, **result)

    def find(self) -> List[Bundle]:
        """Find all bundles.

        Returns
        -------
        list of Bundle
            List of all found bundles.
        """
        path = f"v1/content/{self.content_guid}/bundles"
        url = urls.append(self.config.url, path)
        response = self.session.get(url)
        results = response.json()
        return [
            Bundle(self.config, self.session, **result) for result in results
        ]

    def find_one(self) -> Bundle | None:
        """Find a bundle.

        Returns
        -------
        Bundle | None
            The first found bundle | None if no bundles are found.
        """
        bundles = self.find()
        return next(iter(bundles), None)

    def get(self, id: str) -> Bundle:
        """Get a bundle.

        Parameters
        ----------
        id : str
            Identifier of the bundle to retrieve.

        Returns
        -------
        Bundle
            The bundle with the specified ID.
        """
        path = f"v1/content/{self.content_guid}/bundles/{id}"
        url = urls.append(self.config.url, path)
        response = self.session.get(url)
        result = response.json()
        return Bundle(self.config, self.session, **result)



"""Content resources."""

from __future__ import annotations

from typing import List, Optional, overload


from requests import Session

from . import tasks, urls

from .config import Config
from .bundles import Bundles
from .permissions import Permissions
from .resources import Resources, Resource
from .users import Users


class ContentItemOwner(Resource):
    """Owner information."""

    @property
    def guid(self) -> str:
        return self.get("guid")  # type: ignore

    @property
    def username(self) -> str:
        return self.get("username")  # type: ignore

    @property
    def first_name(self) -> Optional[str]:
        return self.get("first_name")  # type: ignore

    @property
    def last_name(self) -> Optional[str]:
        return self.get("last_name")  # type: ignore


class ContentItem(Resource):
    """Content item resource.

    Attributes
    ----------
    bundles : Bundles
        Bundles resource for the content item.
    permissions : Permissions
        Permissions resource for the content item.
    id : str
        Unique identifier of the content item.
    guid : str
        Globally unique identifier of the content item.
    name : str
        Name of the content item.
    title : Optional[str]
        Title of the content item.
    description : str
        Description of the content item.
    access_type : str
        Access type of the content item.
    connection_timeout : Optional[int]
        Connection timeout setting for the content item.
    read_timeout : Optional[int]
        Read timeout setting for the content item.
    init_timeout : Optional[int]
        Initialization timeout setting for the content item.
    idle_timeout : Optional[int]
        Idle timeout setting for the content item.
    max_processes : Optional[int]
        Maximum number of processes allowed for the content item.
    min_processes : Optional[int]
        Minimum number of processes required for the content item.
    max_conns_per_process : Optional[int]
        Maximum number of connections per process for the content item.
    load_factor : Optional[float]
        Load factor for the content item.
    cpu_request : Optional[float]
        CPU request for the content item.
    cpu_limit : Optional[float]
        CPU limit for the content item.
    memory_request : Optional[int]
        Memory request for the content item.
    memory_limit : Optional[int]
        Memory limit for the content item.
    amd_gpu_limit : Optional[int]
        AMD GPU limit for the content item.
    nvidia_gpu_limit : Optional[int]
        NVIDIA GPU limit for the content item.
    created_time : str
        Creation time of the content item.
    last_deployed_time : str
        Last deployment time of the content item.
    bundle_id : Optional[str]
        Bundle ID associated with the content item.
    app_mode : str
        Application mode of the content item.
    content_category : Optional[str]
        Content category of the content item.
    parameterized : bool
        Indicates if the content item is parameterized.
    cluster_name : Optional[str]
        Name of the cluster associated with the content item.
    image_name : Optional[str]
        Name of the image associated with the content item.
    default_image_name : Optional[str]
        Default image name for the content item.
    default_r_environment_management : Optional[bool]
        Indicates if R environment management is enabled by default.
    default_py_environment_management : Optional[bool]
        Indicates if Python environment management is enabled by default.
    service_account_name : Optional[str]
        Name of the service account associated with the content item.
    r_version : Optional[str]
        R version used by the content item.
    r_environment_management : Optional[bool]
        Indicates if R environment management is enabled.
    py_version : Optional[str]
        Python version used by the content item.
    py_environment_management : Optional[bool]
        Indicates if Python environment management is enabled.
    quarto_version : Optional[str]
        Quarto version used by the content item.
    run_as : Optional[str]
        User to run the content item as.
    run_as_current_user : bool
        Indicates if the content item runs as the current user.
    owner_guid : str
        GUID of the owner of the content item.
    owner : ContentItemOwner
        Owner information of the content item.
    content_url : str
        URL of the content item.
    dashboard_url : str
        Dashboard URL of the content item.
    app_role : str
        Application role of the content item.
    tags : List[dict]
        Tags associated with the content item.
    """

    # Relationships

    @property
    def bundles(self) -> Bundles:
        return Bundles(self.config, self.session, self.guid)

    @property
    def permissions(self) -> Permissions:
        return Permissions(self.config, self.session, self.guid)

    @property
    def owner(self) -> ContentItemOwner:
        if "owner" not in self:
            # It is possible to get a content item that does not contain owner.
            # "owner" is an optional additional request param.
            # If it's not included, we can retrieve the information by `owner_guid`
            self["owner"] = Users(self.config, self.session).get(
                self.owner_guid
            )
        return ContentItemOwner(self.config, self.session, **self["owner"])

    # Properties

    @property
    def id(self) -> str:
        return self.get("id")  # type: ignore

    @property
    def guid(self) -> str:
        return self.get("guid")  # type: ignore

    @property
    def name(self) -> str:
        return self.get("name")  # type: ignore

    @property
    def title(self) -> Optional[str]:
        return self.get("title")  # type: ignore

    @property
    def description(self) -> str:
        return self.get("description")  # type: ignore

    @property
    def access_type(self) -> str:
        return self.get("access_type")  # type: ignore

    @property
    def connection_timeout(self) -> Optional[int]:
        return self.get("connection_timeout")  # type: ignore

    @property
    def read_timeout(self) -> Optional[int]:
        return self.get("read_timeout")  # type: ignore

    @property
    def init_timeout(self) -> Optional[int]:
        return self.get("init_timeout")  # type: ignore

    @property
    def idle_timeout(self) -> Optional[int]:
        return self.get("idle_timeout")  # type: ignore

    @property
    def max_processes(self) -> Optional[int]:
        return self.get("max_processes")  # type: ignore

    @property
    def min_processes(self) -> Optional[int]:
        return self.get("min_processes")  # type: ignore

    @property
    def max_conns_per_process(self) -> Optional[int]:
        return self.get("max_conns_per_process")  # type: ignore

    @property
    def load_factor(self) -> Optional[float]:
        return self.get("load_factor")  # type: ignore

    @property
    def cpu_request(self) -> Optional[float]:
        return self.get("cpu_request")  # type: ignore

    @property
    def cpu_limit(self) -> Optional[float]:
        return self.get("cpu_limit")  # type: ignore

    @property
    def memory_request(self) -> Optional[int]:
        return self.get("memory_request")  # type: ignore

    @property
    def memory_limit(self) -> Optional[int]:
        return self.get("memory_limit")  # type: ignore

    @property
    def amd_gpu_limit(self) -> Optional[int]:
        return self.get("amd_gpu_limit")  # type: ignore

    @property
    def nvidia_gpu_limit(self) -> Optional[int]:
        return self.get("nvidia_gpu_limit")  # type: ignore

    @property
    def created_time(self) -> str:
        return self.get("created_time")  # type: ignore

    @property
    def last_deployed_time(self) -> str:
        return self.get("last_deployed_time")  # type: ignore

    @property
    def bundle_id(self) -> Optional[str]:
        return self.get("bundle_id")  # type: ignore

    @property
    def app_mode(self) -> str:
        return self.get("app_mode")  # type: ignore

    @property
    def content_category(self) -> Optional[str]:
        return self.get("content_category")  # type: ignore

    @property
    def parameterized(self) -> bool:
        return self.get("parameterized")  # type: ignore

    @property
    def cluster_name(self) -> Optional[str]:
        return self.get("cluster_name")  # type: ignore

    @property
    def image_name(self) -> Optional[str]:
        return self.get("image_name")  # type: ignore

    @property
    def default_image_name(self) -> Optional[str]:
        return self.get("default_image_name")  # type: ignore

    @property
    def default_r_environment_management(self) -> Optional[bool]:
        return self.get("default_r_environment_management")  # type: ignore

    @property
    def default_py_environment_management(self) -> Optional[bool]:
        return self.get("default_py_environment_management")  # type: ignore

    @property
    def service_account_name(self) -> Optional[str]:
        return self.get("service_account_name")  # type: ignore

    @property
    def r_version(self) -> Optional[str]:
        return self.get("r_version")  # type: ignore

    @property
    def r_environment_management(self) -> Optional[bool]:
        return self.get("r_environment_management")  # type: ignore

    @property
    def py_version(self) -> Optional[str]:
        return self.get("py_version")  # type: ignore

    @property
    def py_environment_management(self) -> Optional[bool]:
        return self.get("py_environment_management")  # type: ignore

    @property
    def quarto_version(self) -> Optional[str]:
        return self.get("quarto_version")  # type: ignore

    @property
    def run_as(self) -> Optional[str]:
        return self.get("run_as")  # type: ignore

    @property
    def run_as_current_user(self) -> bool:
        return self.get("run_as_current_user")  # type: ignore

    @property
    def owner_guid(self) -> str:
        return self.get("owner_guid")  # type: ignore

    @property
    def content_url(self) -> str:
        return self.get("content_url")  # type: ignore

    @property
    def dashboard_url(self) -> str:
        return self.get("dashboard_url")  # type: ignore

    @property
    def app_role(self) -> str:
        return self.get("app_role")  # type: ignore

    @property
    def tags(self) -> List[dict]:
        return self.get("tags", [])

    # CRUD Methods

    def delete(self) -> None:
        """Delete the content item."""
        path = f"v1/content/{self.guid}"
        url = urls.append(self.config.url, path)
        self.session.delete(url)

    def deploy(self) -> tasks.Task:
        """Deploy the content.

        Spawns an asynchronous task, which activates the latest bundle.

        Returns
        -------
        tasks.Task
            The task for the deployment.

        Examples
        --------
        >>> task = content.deploy()
        >>> task.wait_for()
        None
        """
        path = f"v1/content/{self.guid}/deploy"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, json={"bundle_id": None})
        result = response.json()
        ts = tasks.Tasks(self.config, self.session)
        return ts.get(result["task_id"])

    @overload
    def update(
        self,
        name: str = ...,
        title: Optional[str] = ...,
        description: str = ...,
        access_type: str = ...,
        owner_guid: Optional[str] = ...,
        connection_timeout: Optional[int] = ...,
        read_timeout: Optional[int] = ...,
        init_timeout: Optional[int] = ...,
        idle_timeout: Optional[int] = ...,
        max_processes: Optional[int] = ...,
        min_processes: Optional[int] = ...,
        max_conns_per_process: Optional[int] = ...,
        load_factor: Optional[float] = ...,
        cpu_request: Optional[float] = ...,
        cpu_limit: Optional[float] = ...,
        memory_request: Optional[int] = ...,
        memory_limit: Optional[int] = ...,
        amd_gpu_limit: Optional[int] = ...,
        nvidia_gpu_limit: Optional[int] = ...,
        run_as: Optional[str] = ...,
        run_as_current_user: Optional[bool] = ...,
        default_image_name: Optional[str] = ...,
        default_r_environment_management: Optional[bool] = ...,
        default_py_environment_management: Optional[bool] = ...,
        service_account_name: Optional[str] = ...,
    ) -> None:
        """Update the content item.

        Parameters
        ----------
        name : str, optional
        title : Optional[str], optional
        description : str, optional
        access_type : str, optional
        owner_guid : Optional[str], optional
        connection_timeout : Optional[int], optional
        read_timeout : Optional[int], optional
        init_timeout : Optional[int], optional
        idle_timeout : Optional[int], optional
        max_processes : Optional[int], optional
        min_processes : Optional[int], optional
        max_conns_per_process : Optional[int], optional
        load_factor : Optional[float], optional
        cpu_request : Optional[float], optional
        cpu_limit : Optional[float], optional
        memory_request : Optional[int], optional
        memory_limit : Optional[int], optional
        amd_gpu_limit : Optional[int], optional
        nvidia_gpu_limit : Optional[int], optional
        run_as : Optional[str], optional
        run_as_current_user : Optional[bool], optional
        default_image_name : Optional[str], optional
        default_r_environment_management : Optional[bool], optional
        default_py_environment_management : Optional[bool], optional
        service_account_name : Optional[str], optional
        """
        ...

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the content item."""
        ...

    def update(self, *args, **kwargs) -> None:
        """Update the content item."""
        body = dict(*args, **kwargs)
        url = urls.append(self.config.url, f"v1/content/{self.guid}")
        response = self.session.patch(url, json=body)
        super().update(**response.json())


class Content(Resources):
    """Content resource."""

    def __init__(self, config: Config, session: Session) -> None:
        self.url = urls.append(config.url, "v1/content")
        self.config = config
        self.session = session

    def count(self) -> int:
        """Count the number of content items.

        Returns
        -------
        int
        """
        results = self.session.get(self.url).json()
        return len(results)

    @overload
    def create(
        self,
        name: str = ...,
        title: Optional[str] = ...,
        description: str = ...,
        access_type: str = ...,
        connection_timeout: Optional[int] = ...,
        read_timeout: Optional[int] = ...,
        init_timeout: Optional[int] = ...,
        idle_timeout: Optional[int] = ...,
        max_processes: Optional[int] = ...,
        min_processes: Optional[int] = ...,
        max_conns_per_process: Optional[int] = ...,
        load_factor: Optional[float] = ...,
        cpu_request: Optional[float] = ...,
        cpu_limit: Optional[float] = ...,
        memory_request: Optional[int] = ...,
        memory_limit: Optional[int] = ...,
        amd_gpu_limit: Optional[int] = ...,
        nvidia_gpu_limit: Optional[int] = ...,
        run_as: Optional[str] = ...,
        run_as_current_user: Optional[bool] = ...,
        default_image_name: Optional[str] = ...,
        default_r_environment_management: Optional[bool] = ...,
        default_py_environment_management: Optional[bool] = ...,
        service_account_name: Optional[str] = ...,
    ) -> ContentItem:
        """Create a content item.

        Parameters
        ----------
        name : str, optional
        title : Optional[str], optional
        description : str, optional
        access_type : str, optional
        connection_timeout : Optional[int], optional
        read_timeout : Optional[int], optional
        init_timeout : Optional[int], optional
        idle_timeout : Optional[int], optional
        max_processes : Optional[int], optional
        min_processes : Optional[int], optional
        max_conns_per_process : Optional[int], optional
        load_factor : Optional[float], optional
        cpu_request : Optional[float], optional
        cpu_limit : Optional[float], optional
        memory_request : Optional[int], optional
        memory_limit : Optional[int], optional
        amd_gpu_limit : Optional[int], optional
        nvidia_gpu_limit : Optional[int], optional
        run_as : Optional[str], optional
        run_as_current_user : Optional[bool], optional
        default_image_name : Optional[str], optional
        default_r_environment_management : Optional[bool], optional
        default_py_environment_management : Optional[bool], optional
        service_account_name : Optional[str], optional

        Returns
        -------
        ContentItem
        """
        ...

    @overload
    def create(self, *args, **kwargs) -> ContentItem:
        """Create a content item.

        Returns
        -------
        ContentItem
        """
        ...

    def create(self, *args, **kwargs) -> ContentItem:
        """Create a content item.

        Returns
        -------
        ContentItem
        """
        body = dict(*args, **kwargs)
        path = "v1/content"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, json=body)
        return ContentItem(self.config, self.session, **response.json())

    @overload
    def find(
        self,
        owner_guid: str = ...,
        name: str = ...,
        include: Optional[str] = "owner,tags",
    ) -> List[ContentItem]:
        """Find content items.

        Parameters
        ----------
        owner_guid : str, optional
            The owner's unique identifier, by default ...
        name : str, optional
            The simple URL friendly name, by default ...
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        List[ContentItem]
        """
        ...

    @overload
    def find(
        self, *args, include: Optional[str] = "owner,tags", **kwargs
    ) -> List[ContentItem]:
        """Find content items.

        Parameters
        ----------
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        List[ContentItem]
        """
        ...

    def find(
        self, *args, include: Optional[str] = "owner,tags", **kwargs
    ) -> List[ContentItem]:
        """Find content items.

        Parameters
        ----------
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        List[ContentItem]
        """
        params = dict(*args, include=include, **kwargs)
        response = self.session.get(self.url, params=params)
        results = response.json()
        items = (
            ContentItem(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        )
        return [item for item in items]

    @overload
    def find_one(
        self,
        owner_guid: str = ...,
        name: str = ...,
        include: Optional[str] = "owner,tags",
    ) -> ContentItem | None:
        """Find content items.

        Parameters
        ----------
        owner_guid : str, optional
            The owner's unique identifier, by default ...
        name : str, optional
            The simple URL friendly name, by default ...
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        ContentItem | None
        """
        ...

    @overload
    def find_one(
        self, *args, include: Optional[str] = "owner,tags", **kwargs
    ) -> ContentItem | None:
        """Find content items.

        Parameters
        ----------
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        ContentItem | None
        """
        ...

    def find_one(
        self, *args, include: Optional[str] = "owner,tags", **kwargs
    ) -> ContentItem | None:
        """Find content items.

        Parameters
        ----------
        include : Optional[str], optional
            Comma separated list of details to include in the response, allows 'owner' and 'tags', by default 'owner,tags'

        Returns
        -------
        ContentItem | None
        """
        items = self.find(*args, include=include, **kwargs)
        return next(iter(items), None)

    def get(self, guid: str) -> ContentItem:
        """Get a content item.

        Parameters
        ----------
        guid : str

        Returns
        -------
        ContentItem
        """
        url = urls.append(self.url, guid)
        response = self.session.get(url)
        return ContentItem(self.config, self.session, **response.json())



"""Group resources."""

from __future__ import annotations
from typing import List, overload

import requests

from . import me, urls

from .config import Config
from .paginator import Paginator
from .resources import Resource, Resources


class Group(Resource):
    """Group resource.

    Attributes
    ----------
    guid : str
    name: str
    owner_guid: str
    """

    @property
    def guid(self) -> str:
        return self.get("guid")  # type: ignore

    @property
    def name(self) -> str:
        return self.get("name")  # type: ignore

    @property
    def owner_guid(self) -> str:
        return self.get("owner_guid")  # type: ignore

    # CRUD Methods

    def delete(self) -> None:
        """Delete the group."""
        path = f"v1/groups/{self.guid}"
        url = urls.append(self.config.url, path)
        self.session.delete(url)


class Groups(Resources):
    """Groups resource."""

    def __init__(self, config: Config, session: requests.Session) -> None:
        self.config = config
        self.session = session

    @overload
    def create(self, name: str, unique_id: str | None) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group
        """
        ...

    @overload
    def create(self, *args, **kwargs) -> Group:
        """Create a group.

        Returns
        -------
        Group
        """
        ...

    def create(self, *args, **kwargs) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group
        """
        ...
        body = dict(*args, **kwargs)
        path = "v1/groups"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, json=body)
        return Group(self.config, self.session, **response.json())

    @overload
    def find(
        self,
        prefix: str = ...,
    ) -> List[Group]: ...

    @overload
    def find(self, *args, **kwargs) -> List[Group]: ...

    def find(self, *args, **kwargs):
        """Find groups.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        List[Group]
        """
        params = dict(*args, **kwargs)
        path = "v1/groups"
        url = urls.append(self.config.url, path)
        paginator = Paginator(self.session, url, params=params)
        results = paginator.fetch_results()
        return [
            Group(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        ]

    @overload
    def find_one(
        self,
        prefix: str = ...,
    ) -> Group | None: ...

    @overload
    def find_one(self, *args, **kwargs) -> Group | None: ...

    def find_one(self, *args, **kwargs) -> Group | None:
        """Find one group.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        Group | None
        """
        params = dict(*args, **kwargs)
        path = "v1/groups"
        url = urls.append(self.config.url, path)
        paginator = Paginator(self.session, url, params=params)
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        groups = (
            Group(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        )
        return next(groups, None)

    def get(self, guid: str) -> Group:
        """Get group.

        Parameters
        ----------
        guid : str

        Returns
        -------
        Group
        """
        url = urls.append(self.config.url, f"v1/groups/{guid}")
        response = self.session.get(url)
        return Group(
            config=self.config,
            session=self.session,
            **response.json(),
        )

    def count(self) -> int:
        """Count the number of groups.

        Returns
        -------
        int
        """
        path = "v1/groups"
        url = urls.append(self.config.url, path)
        response: requests.Response = self.session.get(
            url, params={"page_size": 1}
        )
        result: dict = response.json()
        return result["total"]



import io

import pytest
import requests
import responses

from responses import matchers
from unittest import mock

from posit.connect import Client
from posit.connect.config import Config
from posit.connect.bundles import Bundle

from .api import load_mock, get_path  # type: ignore


class TestBundleProperties:
    def setup_class(cls):
        config = Config(api_key="12345", url="https://connect.example/")
        session = requests.Session()
        cls.bundle = Bundle(
            config,
            session,
            **load_mock(
                f"v1/content/f2f37341-e21d-3d80-c698-a935ad614066/bundles/101.json"
            ),
        )

    def test_id(self):
        assert self.bundle.id == "101"

    def test_content_guid(self):
        assert (
            self.bundle.content_guid == "f2f37341-e21d-3d80-c698-a935ad614066"
        )

    def test_created_time(self):
        assert self.bundle.created_time == "2006-01-02T15:04:05Z07:00"

    def test_cluster_name(self):
        assert self.bundle.cluster_name == "Local"

    def test_image_name(self):
        assert self.bundle.image_name == "Local"

    def test_r_version(self):
        assert self.bundle.r_version == "3.5.1"

    def test_r_environment_management(self):
        assert self.bundle.r_environment_management == True

    def test_py_version(self):
        assert self.bundle.py_version == "3.8.2"

    def test_py_environment_management(self):
        assert self.bundle.py_environment_management == True

    def test_quarto_version(self):
        assert self.bundle.quarto_version == "0.2.22"

    def test_active(self):
        assert self.bundle.active == False

    def test_size(self):
        assert self.bundle.size == 1000000

    def test_metadata_source(self):
        assert self.bundle.metadata.source == "string"

    def test_metadata_source_repo(self):
        assert self.bundle.metadata.source_repo == "string"

    def test_metadata_source_branch(self):
        assert self.bundle.metadata.source_branch == "string"

    def test_metadata_source_commit(self):
        assert self.bundle.metadata.source_commit == "string"

    def test_metadata_archive_md5(self):
        assert (
            self.bundle.metadata.archive_md5
            == "37324238a80595c453c706b22adb83d3"
        )

    def test_metadata_archive_sha1(self):
        assert (
            self.bundle.metadata.archive_sha1
            == "a2f7d13d87657df599aeeabdb70194d508cfa92f"
        )


class TestBundleDelete:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        mock_bundle_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
        )

        # setup
        c = Client("12345", "https://connect.example")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        bundle.delete()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_delete.call_count == 1


class TestBundleDeploy:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        task_id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        mock_bundle_deploy = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/deploy",
            match=[matchers.json_params_matcher({"bundle_id": bundle_id})],
            json={"task_id": task_id},
        )

        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{task_id}",
            json=load_mock(f"v1/tasks/{task_id}.json"),
        )

        # setup
        c = Client("12345", "https://connect.example")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        task = bundle.deploy()

        # assert
        task.id == task_id
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_deploy.call_count == 1
        assert mock_tasks_get.call_count == 1


class TestBundleDownload:
    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @responses.activate
    def test_output_as_str(self, mock_file: mock.MagicMock):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz"
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("12345", "https://connect.example")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        bundle.download("pathname")

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_download.call_count == 1
        mock_file.assert_called_once_with("pathname", "wb")

    @responses.activate
    def test_output_as_io(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz"
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("12345", "https://connect.example")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        file = io.BytesIO()
        buffer = io.BufferedWriter(file)
        bundle.download(buffer)
        buffer.seek(0)

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_download.call_count == 1
        assert file.read() == path.read_bytes()

    @responses.activate
    def test_invalid_arguments(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz"
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("12345", "https://connect.example")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        with pytest.raises(TypeError):
            bundle.download(None)

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1


class TestBundlesCreate:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        pathname = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz"
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_post = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        # setup
        c = Client("12345", "https://connect.example")
        content = c.content.get(content_guid)

        # invoke
        data = pathname.read_bytes()
        bundle = content.bundles.create(data)

        # # assert
        assert bundle.id == "101"
        assert mock_content_get.call_count == 1
        assert mock_bundle_post.call_count == 1

    @responses.activate
    def test_kwargs_pathname(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        pathname = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz"
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_post = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        # setup
        c = Client("12345", "https://connect.example")
        content = c.content.get(content_guid)

        # invoke
        pathname = str(pathname.absolute())
        bundle = content.bundles.create(pathname)

        # # assert
        assert bundle.id == "101"
        assert mock_content_get.call_count == 1
        assert mock_bundle_post.call_count == 1

    @responses.activate
    def test_invalid_arguments(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        # setup
        c = Client("12345", "https://connect.example")
        content = c.content.get(content_guid)

        # invoke
        with pytest.raises(TypeError):
            content.bundles.create(None)


class TestBundlesFind:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundles_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles.json"),
        )

        # setup
        c = Client("12345", "https://connect.example")

        # invoke
        bundles = c.content.get(content_guid).bundles.find()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundles_get.call_count == 1
        assert len(bundles) == 1
        assert bundles[0].id == "101"


class TestBundlesFindOne:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundles_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles.json"),
        )

        # setup
        c = Client("12345", "https://connect.example")

        # invoke
        bundle = c.content.get(content_guid).bundles.find_one()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundles_get.call_count == 1
        assert bundle.id == "101"


class TestBundlesGet:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(
                f"v1/content/{content_guid}/bundles/{bundle_id}.json"
            ),
        )

        # setup
        c = Client("12345", "https://connect.example")

        # invoke
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert bundle.id == bundle_id



from unittest.mock import Mock

import pytest
import requests
import responses


from posit.connect.client import Client
from posit.connect.config import Config
from posit.connect.groups import Group

from .api import load_mock  # type: ignore

session = Mock()
url = Mock()


class TestGroupAttributes:
    @classmethod
    def setup_class(cls):
        guid = "6f300623-1e0c-48e6-a473-ddf630c0c0c3"
        config = Config(api_key="12345", url="https://connect.example.com/")
        session = requests.Session()
        fake_item = load_mock(f"v1/groups/{guid}.json")
        cls.item = Group(config, session, **fake_item)

    def test_guid(self):
        assert self.item.guid == "6f300623-1e0c-48e6-a473-ddf630c0c0c3"

    def test_name(self):
        assert self.item.name == "Friends"

    def test_owner_guid(self):
        assert self.item.owner_guid == "20a79ce3-6e87-4522-9faf-be24228800a4"



class ClientError(Exception):
    def __init__(
        self,
        error_code: int,
        error_message: str,
        http_status: int,
        http_message: str,
    ):
        self.error_code = error_code
        self.error_message = error_message
        self.http_status = http_status
        self.http_message = http_message
        super().__init__(
            f"{error_message} (Error Code: {error_code}, HTTP Status: {http_status} {http_message})"
        )



import requests
import responses

from responses import matchers

from posit import connect
from posit.connect import tasks

from .api import load_mock  # type: ignore


class TestTaskAttributes:
    def setup_class(cls):
        cls.task = tasks.Task(
            None,
            None,
            **load_mock("v1/tasks/jXhOhdm5OOSkGhJw.json"),
        )

    def test_id(self):
        assert self.task.id == "jXhOhdm5OOSkGhJw"

    def test_is_finished(self):
        assert self.task.is_finished

    def test_output(self):
        assert self.task.output == [
            "Building static content...",
            "Launching static content...",
        ]

    def test_error_code(self):
        assert self.task.error_code == 1

    def test_error_message(self):
        assert (
            self.task.error_message
            == "Unable to render: Rendering exited abnormally: exit status 1"
        )

    def test_result(self):
        assert self.task.result is None


class TestTaskUpdate:
    @responses.activate
    def test(self):
        id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_tasks_get = [0] * 2
        mock_tasks_get[0] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": False},
        )

        mock_tasks_get[1] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": True},
        )

        # setup
        c = connect.Client("12345", "https://connect.example")
        task = c.tasks.get(id)
        assert not task.is_finished

        # invoke
        task.update()

        # assert
        assert task.is_finished
        assert mock_tasks_get[0].call_count == 1
        assert mock_tasks_get[1].call_count == 1

    @responses.activate
    def test_with_params(self):
        id = "jXhOhdm5OOSkGhJw"
        params = {"first": 10, "wait": 10}

        # behavior
        mock_tasks_get = [0] * 2
        mock_tasks_get[0] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": False},
        )

        mock_tasks_get[1] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": True},
            match=[matchers.query_param_matcher(params)],
        )

        # setup
        c = connect.Client("12345", "https://connect.example")
        task = c.tasks.get(id)
        assert not task.is_finished

        # invoke
        task.update(**params)

        # assert
        assert task.is_finished
        assert mock_tasks_get[0].call_count == 1
        assert mock_tasks_get[1].call_count == 1


class TestTaskWaitFor:
    @responses.activate
    def test(self):
        id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_tasks_get = [0] * 2
        mock_tasks_get[0] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": False},
        )

        mock_tasks_get[1] = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": True},
        )

        # setup
        c = connect.Client("12345", "https://connect.example")
        task = c.tasks.get(id)
        assert not task.is_finished

        # invoke
        task.wait_for()

        # assert
        assert task.is_finished
        assert mock_tasks_get[0].call_count == 1
        assert mock_tasks_get[1].call_count == 1


class TestTasksGet:
    @responses.activate
    def test(self):
        id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{id}",
            json={**load_mock(f"v1/tasks/{id}.json"), "finished": False},
        )

        # setup
        c = connect.Client("12345", "https://connect.example")

        # invoke
        task = c.tasks.get(id)

        # assert
        assert task.id == id
        assert mock_tasks_get.call_count == 1



from posit import connect


class TestContent:
    def setup_class(cls):
        cls.client = connect.Client()
        cls.item = cls.client.content.create(
            name="Sample",
            description="Simple sample content for testing",
            access_type="acl",
        )

    def test_count(self):
        assert self.client.content.count() == 1

    def test_get(self):
        assert self.client.content.get(self.item.guid) == self.item

    def test_find(self):
        assert self.client.content.find()

    def test_find_one(self):
        assert self.client.content.find_one()

    def test_content_item_owner(self):
        item = self.client.content.find_one(include=None)
        owner = item.owner
        assert owner.guid == self.client.me.guid

    def test_content_item_owner_from_include(self):
        item = self.client.content.find_one(include="owner")
        owner = item.owner
        assert owner.guid == self.client.me.guid



from __future__ import annotations

from dataclasses import dataclass, make_dataclass
from typing import Generator, List

import requests

# The maximum page size supported by the API.
_MAX_PAGE_SIZE = 500


@dataclass
class CursorPage:
    paging: dict
    results: List[dict]


class CursorPaginator:
    def __init__(
        self, session: requests.Session, url: str, params: dict = {}
    ) -> None:
        self.session = session
        self.url = url
        self.params = params

    def fetch_results(self) -> List[dict]:
        """Fetch results.

        Collects all results from all pages.

        Returns
        -------
        List[dict]
            A coalesced list of all results.
        """
        results = []
        for page in self.fetch_pages():
            results.extend(page.results)
        return results

    def fetch_pages(self) -> Generator[CursorPage, None, None]:
        """Fetch pages.

        Yields
        ------
        Generator[Page, None, None]
        """
        next = None
        while True:
            page = self.fetch_page(next)
            yield page
            cursors: dict = page.paging.get("cursors", {})
            next = cursors.get("next")
            if not next:
                # stop if a next page is not defined
                return

    def fetch_page(self, next: str | None = None) -> CursorPage:
        """Fetch a page.

        Parameters
        ----------
        next : str | None, optional
            the next page identifier or None to fetch the first page, by default None

        Returns
        -------
        Page
        """
        params = {
            **self.params,
            "next": next,
            "limit": _MAX_PAGE_SIZE,
        }
        response = self.session.get(self.url, params=params)
        return CursorPage(**response.json())



"""Permission resources."""

from __future__ import annotations

from typing import List, overload

from requests.sessions import Session as Session

from . import urls

from .config import Config
from .resources import Resource, Resources


class Permission(Resource):
    @property
    def id(self) -> str:
        return self.get("id")  # type: ignore

    @property
    def content_guid(self) -> str:
        return self.get("content_guid")  # type: ignore

    @property
    def principal_guid(self) -> str:
        return self.get("principal_guid")  # type: ignore

    @property
    def principal_type(self) -> str:
        return self.get("principal_type")  # type: ignore

    @property
    def role(self) -> str:
        return self.get("role")  # type: ignore

    def delete(self) -> None:
        """Delete the permission."""
        path = f"v1/content/{self.content_guid}/permissions/{self.id}"
        url = urls.append(self.config.url, path)
        self.session.delete(url)

    @overload
    def update(self, role: str) -> None:
        """Update the permission.

        Parameters
        ----------
        role : str
            The principal role.
        """
        ...

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the permission."""
        ...

    def update(self, *args, **kwargs) -> None:
        """Update the permission."""
        body = {
            "principal_guid": self.principal_guid,
            "principal_type": self.principal_type,
            "role": self.role,
        }
        body.update(*args, **kwargs)
        path = f"v1/content/{self.content_guid}/permissions/{self.id}"
        url = urls.append(self.config.url, path)
        response = self.session.put(
            url,
            json=body,
        )
        super().update(**response.json())


class Permissions(Resources):
    def __init__(
        self, config: Config, session: Session, content_guid: str
    ) -> None:
        super().__init__(config, session)
        self.content_guid = content_guid

    def count(self) -> int:
        """Count the number of permissions.

        Returns
        -------
        int
        """
        return len(self.find())

    @overload
    def create(
        self, principal_guid: str, principal_type: str, role: str
    ) -> Permission:
        """Create a permission.

        Parameters
        ----------
        principal_guid : str
        principal_type : str
        role : str

        Returns
        -------
        Permission
        """
        ...

    @overload
    def create(self, *args, **kwargs) -> Permission:
        """Create a permission.

        Returns
        -------
        Permission
        """
        ...

    def create(self, *args, **kwargs) -> Permission:
        """Create a permission.

        Returns
        -------
        Permission
        """
        ...
        body = dict(*args, **kwargs)
        path = f"v1/content/{self.content_guid}/permissions"
        url = urls.append(self.config.url, path)
        response = self.session.post(url, json=body)
        return Permission(self.config, self.session, **response.json())

    def find(self, *args, **kwargs) -> List[Permission]:
        """Find permissions.

        Returns
        -------
        List[Permission]
        """
        body = dict(*args, **kwargs)
        path = f"v1/content/{self.content_guid}/permissions"
        url = urls.append(self.config.url, path)
        response = self.session.get(url, json=body)
        results = response.json()
        return [
            Permission(self.config, self.session, **result)
            for result in results
        ]

    def find_one(self, *args, **kwargs) -> Permission | None:
        """Find a permission.

        Returns
        -------
        Permission | None
        """
        permissions = self.find(*args, **kwargs)
        return next(iter(permissions), None)

    def get(self, id: str) -> Permission:
        """Get a permission.

        Parameters
        ----------
        id : str
            The permission id.

        Returns
        -------
        Permission
        """
        path = f"v1/content/{self.content_guid}/permissions/{id}"
        url = urls.append(self.config.url, path)
        response = self.session.get(url)
        return Permission(self.config, self.session, **response.json())



from __future__ import annotations

import posixpath

from urllib.parse import urlsplit, urlunsplit

Url = str


def create(url: str) -> Url:
    """Create a Url.

    Asserts that the Url is a proper Posit Connect endpoint. The path '__api__' is appended to the Url if it isn't already present.

    Parameters
    ----------
    url : str
        The original Url.

    Returns
    -------
    Url
        The validated and formatted Url.

    Raises
    ------
    ValueError
        The Url is missing a scheme.
    ValueError
        The Url is missing a network location (i.e., a domain name).

    Examples
    --------
    >>> urls.create("http://example.com")
    http://example.com/__api__

    >>> urls.create("http://example.com/__api__")
    http://example.com/__api__

    """
    split = urlsplit(url, allow_fragments=False)
    if not split.scheme:
        raise ValueError(
            f"url must specify a scheme (e.g., http://example.com/__api__): {url}"
        )

    if not split.netloc:
        raise ValueError(
            f"url must be absolute (e.g., http://example.com/__api__): {url}"
        )

    url = url.rstrip("/")
    if not url.endswith("__api__"):
        url = append(url, "__api__")

    return url


def append(url: Url, path: str) -> Url:
    """Append a path to a Url.

    Parameters
    ----------
    url : Url
        A valid Url.
    path : str
        A valid Url path.

    Returns
    -------
    Url
        The original Url with the path appended to the end.

    Examples
    --------
    >>> url = urls.create("http://example.com/__api__")
    >>> urls.append(url, "path")
    http://example.com/__api__/path
    """
    # Removes leading '/' from path to avoid double slashes.
    path = path.lstrip("/")
    # Removes trailing '/' from path to avoid double slashes.
    path = path.rstrip("/")
    # See https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlsplit
    split = urlsplit(url, allow_fragments=False)
    # Append the path to unmodified Url path.
    path = posixpath.join(split.path, path)
    # See https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlunsplit
    return urlunsplit((split.scheme, split.netloc, path, split.query, None))



import responses
import requests

from responses import matchers

from posit.connect import Client, config
from posit.connect.metrics import shiny_usage

from ..api import load_mock  # type: ignore


class TestShinyUsageEventAttributes:
    def setup_class(cls):
        cls.event = shiny_usage.ShinyUsageEvent(
            None,
            None,
            **load_mock("v1/instrumentation/shiny/usage?limit=500.json")[
                "results"
            ][0],
        )

    def test_content_guid(self):
        assert (
            self.event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"
        )

    def test_user_guid(self):
        assert self.event.user_guid == "08e3a41d-1f8e-47f2-8855-f05ea3b0d4b2"

    def test_started(self):
        assert self.event.started == "2018-09-15T18:00:00-05:00"

    def test_ended(self):
        assert self.event.ended == "2018-09-15T18:01:00-05:00"

    def test_data_version(self):
        assert self.event.data_version == 1


class TestShinyUsageFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [None] * 2
        mock_get[0] = responses.get(
            f"https://connect.example/__api__/v1/instrumentation/shiny/usage",
            json=load_mock("v1/instrumentation/shiny/usage?limit=500.json"),
            match=[
                matchers.query_param_matcher(
                    {
                        "limit": 500,
                    }
                )
            ],
        )

        mock_get[1] = responses.get(
            f"https://connect.example/__api__/v1/instrumentation/shiny/usage",
            json=load_mock(
                "v1/instrumentation/shiny/usage?limit=500&next=23948901087.json"
            ),
            match=[
                matchers.query_param_matcher(
                    {
                        "next": "23948901087",
                        "limit": 500,
                    }
                )
            ],
        )

        # setup
        c = config.Config("12345", "https://connect.example")
        session = requests.Session()

        # invoke
        events = shiny_usage.ShinyUsage(c, session).find()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 1
        assert len(events) == 1


class TestShinyUsageFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [None] * 2
        mock_get[0] = responses.get(
            f"https://connect.example/__api__/v1/instrumentation/shiny/usage",
            json=load_mock("v1/instrumentation/shiny/usage?limit=500.json"),
            match=[
                matchers.query_param_matcher(
                    {
                        "limit": 500,
                    }
                )
            ],
        )

        mock_get[1] = responses.get(
            f"https://connect.example/__api__/v1/instrumentation/shiny/usage",
            json=load_mock(
                "v1/instrumentation/shiny/usage?limit=500&next=23948901087.json"
            ),
            match=[
                matchers.query_param_matcher(
                    {
                        "next": "23948901087",
                        "limit": 500,
                    }
                )
            ],
        )

        # setup
        c = config.Config("12345", "https://connect.example")
        session = requests.Session()

        # invoke
        event = shiny_usage.ShinyUsage(c, session).find_one()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 0
        assert event
        assert event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"


class TestRenameParams:
    def test_start_to_from(self):
        params = {"start": ...}
        params = shiny_usage.rename_params(params)
        assert "start" not in params
        assert "from" in params

    def test_end_to_to(self):
        params = {"end": ...}
        params = shiny_usage.rename_params(params)
        assert "end" not in params
        assert "to" in params
